# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import datetime

from earthkit.data.core.geography import Geography
from earthkit.data.core.metadata import Metadata
from earthkit.data.indexing.database import GRIB_KEYS_NAMES
from earthkit.data.utils.bbox import BoundingBox
from earthkit.data.utils.projections import Projection


def missing_is_none(x):
    return None if x == 2147483647 else x


class GribFieldGeography(Geography):
    def __init__(self, metadata):
        self.metadata = metadata

    def latitudes(self):
        r"""Return the latitudes of the field.

        Returns
        -------
        ndarray
        """
        return self.metadata._handle.get_latitudes()

    def longitudes(self):
        r"""Return the longitudes of the field.

        Returns
        -------
        ndarray
        """
        return self.metadata._handle.get_longitudes()

    def x(self):
        r"""Return the x coordinates in the field's original CRS.

        Returns
        -------
        ndarray
        """
        grid_type = self.metadata.get("gridType", None)
        if grid_type in ["regular_ll", "reduced_gg", "regular_gg"]:
            return self.longitudes()

    def y(self):
        r"""Return the y coordinates in the field's original CRS.

        Returns
        -------
        ndarray
        """
        grid_type = self.metadata.get("gridType", None)
        if grid_type in ["regular_ll", "reduced_gg", "regular_gg"]:
            return self.latitudes()

    def shape(self):
        r"""Get the shape of the field.

        For structured grids the shape is a tuple in the form of (Nj, Ni) where:

        - ni: the number of gridpoints in i direction (longitude for a regular latitude-longitude grid)
        - nj: the number of gridpoints in j direction (latitude for a regular latitude-longitude grid)

        For other grid types the number of gridpoints is returned as ``(num,)``

        Returns
        -------
        tuple
        """
        Nj = missing_is_none(self.metadata.get("Nj", None))
        Ni = missing_is_none(self.metadata.get("Ni", None))
        if Ni is None or Nj is None:
            n = self.metadata.get("numberOfDataPoints", None)
            return (n,)  # shape must be a tuple
        return (Nj, Ni)

    def _unique_grid_id(self):
        return self.metadata.get("md5GridSection", None)

    def projection(self):
        r"""Return information about the projection.

        Returns
        -------
        :obj:`Projection`

        Examples
        --------
        >>> import earthkit.data
        >>> ds = earthkit.data.from_source("file", "docs/examples/test.grib")
        >>> ds.projection()
        <Projected CRS: +proj=eqc +ellps=WGS84 +a=6378137.0 +lon_0=0.0 +to ...>
        Name: unknown
        Axis Info [cartesian]:
        - E[east]: Easting (unknown)
        - N[north]: Northing (unknown)
        - h[up]: Ellipsoidal height (metre)
        Area of Use:
        - undefined
        Coordinate Operation:
        - name: unknown
        - method: Equidistant Cylindrical
        Datum: Unknown based on WGS 84 ellipsoid
        - Ellipsoid: WGS 84
        - Prime Meridian: Greenwich
        >>> ds.projection().to_proj_string()
        '+proj=eqc +ellps=WGS84 +a=6378137.0 +lon_0=0.0 +to_meter=111319.4907932736 +no_defs +type=crs'
        """
        return Projection.from_proj_string(self.metadata.get("projTargetString", None))

    def bounding_box(self):
        r"""Return the bounding box of the field.

        Returns
        -------
        :obj:`BoundingBox <data.utils.bbox.BoundingBox>`
        """
        return BoundingBox(
            north=self.metadata.get("latitudeOfFirstGridPointInDegrees", None),
            south=self.metadata.get("latitudeOfLastGridPointInDegrees", None),
            west=self.metadata.get("longitudeOfFirstGridPointInDegrees", None),
            east=self.metadata.get("longitudeOfLastGridPointInDegrees", None),
        )


class GribMetadata(Metadata):
    """Represent the metadata of a GRIB field.

    Parameters
    ----------
    handle: :obj:`GribCodesHandle`
        Object representing the ecCodes GRIB handle of the field.


    :obj:`GribMetadata` is created internally by a :obj:`GribField` and we can use
    the field's :meth:`metadata` method to access it.

    >>> ds = earthkit.data.from_source("file", "docs/examples/test4.grib")
    >>> md = ds[0].metadata()
    >>> md["shortName"]
    't'
    >>> md.get("shortName")
    't'
    >>> md.get("nonExistentKey")
    >>> md.get("nonExistentKey", 12)
    12

    """

    LS_KEYS = [
        "centre",
        "shortName",
        "typeOfLevel",
        "level",
        "dataDate",
        "dataTime",
        "stepRange",
        "dataType",
        "number",
        "gridType",
    ]

    DESCRIBE_KEYS = [
        "shortName",
        "typeOfLevel",
        "level",
        "date",
        "time",
        "step",
        "number",
        "paramId",
        "marsClass",
        "marsStream",
        "marsType",
        "experimentVersionNumber",
    ]

    INDEX_KEYS = list(GRIB_KEYS_NAMES)

    NAMESPACES = [
        "default",
        "ls",
        "geography",
        "mars",
        "parameter",
        "statistics",
        "time",
        "vertical",
    ]

    __handle_type = None

    def __init__(self, handle):
        if not isinstance(handle, self._handle_type()):
            raise TypeError(
                f"GribMetadata: expected handle type {self._handle_type()}, got {type(handle)}"
            )
        self._handle = handle
        self._geo = None

    @staticmethod
    def _handle_type():
        """Return the expected handle type. Implemented like this
        to avoid cyclic import
        """
        if GribMetadata.__handle_type is None:
            from earthkit.data.readers.grib.codes import GribCodesHandle

            GribMetadata.__handle_type = GribCodesHandle
        return GribMetadata.__handle_type

    def __len__(self):
        return sum(map(lambda i: 1, self._handle.keys()))

    def __contains__(self, key):
        return self._handle.__contains__(key)

    def keys(self):
        r"""Return a new view of the metadata keys.

        Returns
        -------
        Iterable of str

        """
        return self._handle.keys()

    def items(self):
        r"""Return a new view of the metadata items.

        Returns
        -------
        Iterable of :obj:`(key,value)` pairs

        """
        return self._handle.items()

    def _get_internal_key(self, key, astype=None, default=None, raise_on_missing=False):
        def _key_name(key):
            if key == "param":
                key = "shortName"
            elif key == "_param_id":
                key = "paramId"
            return key

        _kwargs = {}
        if not raise_on_missing:
            _kwargs["default"] = default

        return self._handle.get(_key_name(key), ktype=astype, **_kwargs)

    def override(self, *args, **kwargs):
        r"""Change the metadata values and return a new object.

        :obj:`override` accepts another :obj:`Metadata` or a dict or
        an iterable of key/value pairs (as tuples or other iterables of length two).
        If keyword arguments are specified, the metadata is then updated with those
        key/value pairs.

        Examples
        --------
        >>> other = RawMetadata({"key1": 1, "key2": 2})
        >>> m1 = m.override(other)
        >>> m1 = m.override({"key1": 1, "key2": 2})
        >>> m1 = m.override([("key1", 1), ("key2", 2)])
        >>> m1 = m.override(key1=1, key2=2)
        """
        d = dict(*args, **kwargs)
        handle = self._handle.clone()
        handle.set_multiple(d)
        return GribMetadata(handle)

    def as_namespace(self, namespace=None):
        r"""Return all the keys/values from a namespace.

        Parameters
        ----------
        namespace: str, None
            The :xref:`eccodes_namespace`. earthkit-data also defines the "default" namespace,
            which contains all the GRIB keys that ecCodes can access without specifying a namespace.
            When `namespace` is None or an empty :obj:`str` all the available
            keys/values are returned.

        Returns
        -------
        dict
            All the keys/values from the `namespace`.

        """
        if not isinstance(namespace, str) and namespace is not None:
            raise TypeError("namespace must be a str or None")

        if namespace == "default" or namespace == "":
            namespace = None
        return self._handle.as_namespace(namespace)

    @property
    def geography(self):
        if self._geo is None:
            self._geo = GribFieldGeography(self)
        return self._geo

    def namespaces(self):
        r"""Return the available namespaces.

        Returns
        -------
        list of str
        """
        return self.NAMESPACES

    def datetime(self):
        r"""Return the date and time of the field.

        Returns
        -------
        dict of datatime.datetime
            Dict with items "base_time" and "valid_time".


        >>> import earthkit.data
        >>> ds = earthkit.data.from_source("file", "tests/data/t_time_series.grib")
        >>> ds[4].datetime()
        {'base_time': datetime.datetime(2020, 12, 21, 12, 0),
        'valid_time': datetime.datetime(2020, 12, 21, 18, 0)}

        """
        return {
            "base_time": self._base_datetime(),
            "valid_time": self._valid_datetime(),
        }

    def _base_datetime(self):
        date = self.get("date", None)
        time = self.get("time", None)
        return self._build_datetime(date, time)

    def _valid_datetime(self):
        date = self.get("validityDate", None)
        time = self.get("validityTime", None)
        return self._build_datetime(date, time)

    def _build_datetime(self, date, time):
        return datetime.datetime(
            date // 10000,
            date % 10000 // 100,
            date % 100,
            time // 100,
            time % 100,
        )

    def dump(self, namespace=all, **kwargs):
        r"""Generate dump with all the metadata keys belonging to ``namespace``.

        In a Jupyter notebook it is represented as a tabbed interface.

        Parameters
        ----------
        namespace: :obj:`str`, :obj:`list`, :obj:`tuple`, :obj:`None` or :obj:`all`
            The namespace to dump. Any :xref:`eccodes_namespace` can be used here.
            earthkit-data also defines the "default" namespace, which contains all the GRIB keys
            that ecCodes can access without specifying a namespace. The following `namespace` values
            have a special meaning:

            - :obj:`all`: all the available namespaces will be used (including "default").
            - None or empty str: the "default" namespace will be used.

        **kwargs: dict, optional
            Other keyword arguments used for testing only

        Returns
        -------
        NamespaceDump
            Dict-like object with one item per namespace. In a Jupyter notebook represented
            as a tabbed interface to browse the dump contents.

        Examples
        --------
        :ref:`/examples/grib_metadata.ipynb`

        """
        from earthkit.data.utils.summary import format_namespace_dump

        namespace = self.NAMESPACES if namespace is all else [namespace]
        r = [
            {
                "title": ns if ns else "default",
                "data": self.as_namespace(ns),
                "tooltip": f"Keys in the ecCodes {ns} namespace",
            }
            for ns in namespace
        ]

        return format_namespace_dump(
            r, selected="parameter", details=self.__class__.__name__, **kwargs
        )

    def ls_keys(self):
        r"""Return the keys to be used with the :meth:`ls` method."""
        return self.LS_KEYS

    def describe_keys(self):
        r"""Return the keys to be used with the :meth:`describe` method."""
        return self.DESCRIBE_KEYS

    def index_keys(self):
        r"""Return the keys to be used with the :meth:`indices` method."""
        return self.INDEX_KEYS



#!/usr/bin/env python3

# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import datetime

import pytest

from earthkit.data import from_source
from earthkit.data.testing import earthkit_examples_file


@pytest.mark.parametrize(
    "key,expected_value",
    [
        ("variable", "t"),
        # ("variable:s", "t"),
        # ("variable:str", "t"),
        # ("centre", "ecmf"),
        # ("centre:l", 98),
        ("level", 1000),
        # ("level:l", 1000),
        # ("level:int", 1000),
        (["variable"], ["t"]),
        (["variable", "level"], ["t", 1000]),
        (("variable"), "t"),
        (("variable", "level"), ("t", 1000)),
    ],
)
def test_netcdf_metadata_single_field(key, expected_value):
    f = from_source("file", earthkit_examples_file("tuv_pl.nc"))
    # sn = f.metadata(key)
    # assert sn == [expected_value]
    sn = f[0].metadata(key)
    assert sn == expected_value


def test_netcdf_datetime():
    ds = from_source("file", earthkit_examples_file("test.nc"))

    ref = {
        "base_time": [datetime.datetime(2020, 5, 13, 12)],
        "valid_time": [datetime.datetime(2020, 5, 13, 12)],
    }
    assert ds.datetime() == ref

    ds = from_source(
        "dummy-source",
        kind="netcdf",
        dims=["lat", "lon", "time"],
        variables=["a", "b"],
        coord_values=dict(
            time=[
                datetime.datetime(1990, 1, 1, 12, 0),
                datetime.datetime(1990, 1, 2, 12, 0),
            ]
        ),
    )

    ref = {
        "base_time": [
            datetime.datetime(1990, 1, 1, 12, 0),
            datetime.datetime(1990, 1, 2, 12, 0),
        ],
        "valid_time": [
            datetime.datetime(1990, 1, 1, 12, 0),
            datetime.datetime(1990, 1, 2, 12, 0),
        ],
    }
    assert ds.datetime() == ref


def test_netcdf_valid_datetime():
    ds = from_source("file", earthkit_examples_file("test.nc"))
    assert ds[0].metadata("valid_datetime") == datetime.datetime(2020, 5, 13, 12)


if __name__ == "__main__":
    from earthkit.data.testing import main

    main()



#!/usr/bin/env python3

# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import pytest

from earthkit.data import from_source
from earthkit.data.core.temporary import temp_file
from earthkit.data.testing import earthkit_examples_file


def repeat_list_items(items, count):
    return sum([[x] * count for x in items], [])


def test_grib_from_stream_invalid_args():
    with open(earthkit_examples_file("test6.grib"), "rb") as stream:
        with pytest.raises(TypeError):
            from_source("stream", stream, order_by="level")

    with open(earthkit_examples_file("test6.grib"), "rb") as stream:
        with pytest.raises(TypeError):
            from_source("stream", stream, group_by=1)

    with open(earthkit_examples_file("test6.grib"), "rb") as stream:
        with pytest.raises(TypeError):
            from_source("stream", stream, group_by=["level", 1])

    with open(earthkit_examples_file("test6.grib"), "rb") as stream:
        with pytest.raises(TypeError):
            from_source("stream", stream, group_by="level", batch_size=1)

    with open(earthkit_examples_file("test6.grib"), "rb") as stream:
        with pytest.raises(ValueError):
            from_source("stream", stream, batch_size=-1)


@pytest.mark.parametrize("group_by", ["level", ["level", "gridType"]])
def test_grib_from_stream_group_by(group_by):
    with open(earthkit_examples_file("test6.grib"), "rb") as stream:
        fs = from_source("stream", stream, group_by=group_by)

        # no methods are available
        with pytest.raises(TypeError):
            len(fs)

        ref = [
            [("t", 1000), ("u", 1000), ("v", 1000)],
            [("t", 850), ("u", 850), ("v", 850)],
        ]
        for i, f in enumerate(fs):
            assert len(f) == 3
            assert f.metadata(("param", "level")) == ref[i]

        # stream consumed, no data is available
        assert sum([1 for _ in fs]) == 0


def test_grib_from_stream_single_batch():
    with open(earthkit_examples_file("test6.grib"), "rb") as stream:
        fs = from_source("stream", stream)

        # no methods are available
        with pytest.raises(TypeError):
            len(fs)

        ref = ["t", "u", "v", "t", "u", "v"]
        val = []
        for f in fs:
            v = f.metadata("param")
            val.append(v)

        assert val == ref

        # stream consumed, no data is available
        assert sum([1 for _ in fs]) == 0


def test_grib_from_stream_multi_batch():
    with open(earthkit_examples_file("test6.grib"), "rb") as stream:
        fs = from_source("stream", stream, batch_size=2)

        # no methods are available
        with pytest.raises(TypeError):
            len(fs)

        ref = [["t", "u"], ["v", "t"], ["u", "v"]]
        for i, f in enumerate(fs):
            assert len(f) == 2
            f.metadata("param") == ref[i]

        # stream consumed, no data is available
        assert sum([1 for _ in fs]) == 0


def test_grib_from_stream_in_memory():
    with open(earthkit_examples_file("test6.grib"), "rb") as stream:
        fs = from_source("stream", stream, batch_size=0)

        assert len(fs) == 6

        ref = ["t", "u", "v", "t", "u", "v"]
        val = []

        # iteration
        for f in fs:
            v = f.metadata("param")
            val.append(v)

        assert val == ref, "iteration"

        # metadata
        val = []
        val = fs.metadata("param")
        assert val == ref, "method"


def test_grib_save_when_loaded_from_stream():
    with open(earthkit_examples_file("test6.grib"), "rb") as stream:
        fs = from_source("stream", stream, batch_size=0)
        assert len(fs) == 6
        with temp_file() as tmp:
            fs.save(tmp)
            fs_saved = from_source("file", tmp)
            assert len(fs) == len(fs_saved)


if __name__ == "__main__":
    from earthkit.data.testing import main

    main()



#!/usr/bin/env python3

# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import sys

import pytest

from earthkit.data.arguments.args_kwargs import ArgsKwargs
from earthkit.data.decorators import normalize


def test_normalize_args_kwargs():
    def f(a, *args, x=1, **kwargs):
        return a, args, x, kwargs

    args = [1, 2]
    kwargs = dict(y="Y")
    ak = ArgsKwargs(args, kwargs, f)
    assert ak.positionals_only == []
    assert ak.defaults == {}
    ak.add_default_values_and_kwargs()
    assert ak.positionals_only == []
    assert ak.defaults == dict(x=1)


def test_normalize_kwargs_1():
    class Klass:
        @normalize("param", ["a", "b", "c"])
        def ok(self, param):
            pass

        @normalize("param", ["a", "b", "c"])
        def f(self, **kwargs):
            assert "param" in kwargs

    Klass().ok(param="a")

    Klass().f(param="a")


def test_normalize_kwargs_2():
    @normalize("date", "date-list")
    def f(**kwargs):
        pass

    f()


@pytest.mark.skipif(sys.version_info < (3, 8), reason="Python < 3.8")
def test_normalize_advanced_1():
    exec(
        """
# def f(a,/, b, c=4,*, x=3):
#    return a,b,c,x
# args = ['A']
# kwargs=dict(b=2, c=4)

@normalize("b", ["B", "BB"])
def f(a, /, b, c=4, *, x=3):
    return a, b, c, x

out = f("A", b="B", c=7, x=8)
assert out == ("A", "B", 7, 8), out
"""
    )


@pytest.mark.skipif(sys.version_info < (3, 8), reason="Python < 3.8")
def test_normalize_advanced_2():
    exec(
        """
@normalize("b", ["B", "BB"])
@normalize("a", ["A", "AA"])
def g(a, /, b, c=4, *, x=3):
    return a, b, c, x

out = g("A", b="B", c=7, x=8)
print(out)
assert out == ("A", "B", 7, 8), out
"""
    )


if __name__ == "__main__":
    from earthkit.data.testing import main

    main(__file__)



#!/usr/bin/env python3

# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#


def test_version():
    from earthkit.data import __version__

    assert __version__ is not None


if __name__ == "__main__":
    from earthkit.data.testing import main

    main(__file__)



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#


import glob
import logging
import os

from earthkit.data import from_source
from earthkit.data.core.caching import CACHE
from earthkit.data.readers import reader
from earthkit.data.sources.file_indexed import FileIndexedSource

from . import Source

LOG = logging.getLogger(__name__)


class FileSourceMeta(type(Source), type(os.PathLike)):
    def patch(cls, obj, *args, **kwargs):
        if "reader" in kwargs:
            setattr(obj, "reader", kwargs.pop("reader"))

        return super().patch(obj, *args, **kwargs)


class FileSource(Source, os.PathLike, metaclass=FileSourceMeta):
    _reader_ = None

    def __init__(self, path=None, filter=None, merger=None, **kwargs):
        Source.__init__(self, **kwargs)
        self.path = path
        self.filter = filter
        self.merger = merger

    def mutate(self):
        if isinstance(self.path, (list, tuple)):
            if len(self.path) == 1:
                self.path = self.path[0]
            else:
                return from_source(
                    "multi",
                    [from_source("file", p, **self._kwargs) for p in self.path],
                    filter=self.filter,
                    merger=self.merger,
                )

        # here we must have a file or a directory
        if self._kwargs.get("indexing", False):
            kw = dict(self._kwargs)
            kw.pop("indexing", None)
            return FileIndexedSource(self.path, filter=filter, merger=self.merger, **kw)

        # Give a chance to directories and zip files
        # to return a multi-source
        source = self._reader.mutate_source()
        if source not in (None, self):
            source._parent = self
            return source
        return self

    def ignore(self):
        return self._reader.ignore()

    @classmethod
    def merge(cls, sources):
        from earthkit.data.mergers import merge_by_class

        assert all(isinstance(s, FileSource) for s in sources)
        return merge_by_class([s._reader for s in sources])

    @property
    def _reader(self):
        if self._reader_ is None:
            self._reader_ = reader(self, self.path)
        return self._reader_

    def __iter__(self):
        return iter(self._reader)

    def __len__(self):
        return len(self._reader)

    def __getitem__(self, n):
        return self._reader[n]

    def sel(self, *args, **kwargs):
        return self._reader.sel(*args, **kwargs)

    def isel(self, *args, **kwargs):
        return self._reader.isel(*args, **kwargs)

    def order_by(self, *args, **kwargs):
        return self._reader.order_by(*args, **kwargs)

    def to_xarray(self, **kwargs):
        return self._reader.to_xarray(**kwargs)

    def to_pandas(self, **kwargs):
        LOG.debug("Calling reader.to_pandas %s", self)
        return self._reader.to_pandas(**kwargs)

    def to_numpy(self, **kwargs):
        return self._reader.to_numpy(**kwargs)

    @property
    def values(self):
        return self._reader.values

    def save(self, path):
        return self._reader.save(path)

    def write(self, f):
        return self._reader.write(f)

    def scaled(self, *args, **kwargs):
        return self._reader.scaled(*args, **kwargs)

    def _attributes(self, names):
        return self._reader._attributes(names)

    def __repr__(self):
        path = getattr(self, "path", None)
        if isinstance(path, str):
            cache_dir = CACHE.cache_directory()
            path = path.replace(cache_dir, "CACHE:")
        try:
            reader_class_name = str(self._reader.__class__.__name__)
        except AttributeError as e:
            reader_class_name = str(e)
        except:  # noqa: E722
            reader_class_name = "Unknown"
        return f"{self.__class__.__name__}({path},{reader_class_name})"

    def __fspath__(self):
        return self.path

    def metadata(self, *args, **kwargs):
        return self._reader.metadata(*args, **kwargs)

    def indices(self, *args, **kwargs):
        return self._reader.indices(*args, **kwargs)

    def index(self, *args, **kwargs):
        return self._reader.index(*args, **kwargs)

    def head(self, n=5, **kwargs):
        if n <= 0:
            raise ValueError("head: n must be > 0")
        return self.ls(n=n, **kwargs)

    def tail(self, n=5, **kwargs):
        if n <= 0:
            raise ValueError("n must be > 0")
        return self.ls(n=-n, **kwargs)

    def ls(self, *args, **kwargs):
        return self._reader.ls(*args, **kwargs)

    def describe(self, *args, **kwargs):
        return self._reader.describe(*args, **kwargs)

    def datetime(self, **kwargs):
        return self._reader.datetime(**kwargs)

    def bounding_box(self):
        return self._reader.bounding_box()

    def statistics(self, **kwargs):
        return self._reader.statistics(**kwargs)


class IndexedFileSource(FileSource):
    def mutate(self):
        pass


class File(FileSource):
    def __init__(
        self,
        path,
        expand_user=True,
        expand_vars=False,
        unix_glob=True,
        recursive_glob=True,
        filter=None,
        merger=None,
        **kwargs,
    ):
        if not isinstance(path, (list, tuple)):
            if expand_user:
                path = os.path.expanduser(path)

            if expand_vars:
                path = os.path.expandvars(path)

            if unix_glob and set(path).intersection(set("[]?*")):
                matches = glob.glob(path, recursive=recursive_glob)
                if len(matches) == 1:
                    path = matches[0]
                if len(matches) > 1:
                    path = sorted(matches)

        super().__init__(path, filter, merger, **kwargs)


source = File



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

"""
Module containing methods to transform the inputs of functions based on the function type setting,
common signitures or mapping defined at call time
"""
import inspect
import types
import typing as T
from functools import wraps

from earthkit.data import transform
from earthkit.data.wrappers import Wrapper

try:
    UNION_TYPES = [T.Union, types.UnionType]
except AttributeError:
    # types.UnionType is not in versions of python<3.9
    UNION_TYPES = [
        T.Union,
    ]

EMPTY_TYPES = [inspect._empty]


def _ensure_iterable(input_item):
    """Ensure that an item is iterable"""
    if not isinstance(input_item, (tuple, list, dict)):
        return [input_item]
    return input_item


def _ensure_tuple(input_item):
    """Ensure that an item is a tuple"""
    if not isinstance(input_item, tuple):
        return tuple(_ensure_iterable(input_item))
    return input_item


def transform_function_inputs(
    function: T.Callable,
    kwarg_types: T.Dict[str, T.Any] = {},
    convert_types: T.Union[T.Tuple[T.Any], T.Dict[str, T.Tuple[T.Any]]] = (),
) -> T.Callable:
    """
    Transform the inputs to a function to match the requirements.
    earthkit.data handles the input arg/kwarg format.

    Parameters
    ----------
    function : Callable
        Method to be wrapped
    kwarg_types : Dict[str: type]
        Mapping of accepted object types for each arg/kwarg
    convert_types : Tuple[type]
        List of data-types to try to convert, this can be useful when the function is versitile and can
        accept a large number of data-types, hence only a small number of types should be converted.

    Returns
    -------
    [type]
        [description]
    """

    def _wrapper(kwarg_types, convert_types, *args, **kwargs):
        kwarg_types = {**kwarg_types}
        signature = inspect.signature(function)
        mapping = signature_mapping(signature, kwarg_types)

        # Add args to kwargs for ease of looping:
        arg_names = []
        for arg, name in zip(args, signature.parameters):
            arg_names.append(name)
            kwargs[name] = arg

        convert_kwargs = [k for k in kwargs if k in mapping]
        # Only convert some data-types, this can be used to prevent conversion for for functions which
        #  accept a long-list of formats, e.g. numpy methods can accept xarray, pandas and more

        # Filter for convert_types
        if convert_types:
            # Ensure convert_types is a dictionary
            if not isinstance(convert_types, dict):
                convert_types = {key: convert_types for key in convert_kwargs}

            convert_kwargs = [
                k
                for k in convert_kwargs
                if isinstance(kwargs[k], _ensure_tuple(convert_types.get(k, ())))
            ]

        # transform args/kwargs
        for key in convert_kwargs:
            value = kwargs[key]
            kwarg_types = _ensure_iterable(mapping[key])
            # Transform value if necessary
            if type(value) not in kwarg_types:
                for kwarg_type in kwarg_types:
                    try:
                        kwargs[key] = transform(value, kwarg_type)
                    except Exception:
                        # Transform was not possible, move to next kwarg type.
                        # If no transform is possible, format is unchanged and we rely on function to raise
                        # an Error.
                        continue
                    break

        # Anything that is still a Wrapper object, expand to native data format:
        for k, v in [(_k, _v) for _k, _v in kwargs.items() if isinstance(_v, Wrapper)]:
            try:
                kwargs[k] = v.data
            except Exception:
                pass

        # Extract args from kwargs:
        args = [kwargs.pop(name) for name in arg_names]
        return function(*args, **kwargs)

    @wraps(function)
    def wrapper(*args, _auto_inputs_transform=True, **kwargs):
        if not _auto_inputs_transform:  # Possibility to escape wrapping at call level
            return function(*args, **kwargs)
        return _wrapper(kwarg_types, convert_types, *args, **kwargs)

    return wrapper


def signature_mapping(signature, kwarg_types):
    """
    Map args and kwargs to object types, using hierarchical selection method:
    1. Explicitly defined type
    2. Based on Type setting in function
    3. Do nothing
    """
    mapping = {}
    for key, parameter in signature.parameters.items():
        if key in kwarg_types:
            # 1. Use explicitly defined type
            kwarg_type = kwarg_types.get(key)
        elif parameter.annotation not in EMPTY_TYPES:
            # 2. Use type setting from function
            kwarg_type = parameter.annotation
            if T.get_origin(kwarg_type) in UNION_TYPES:
                # Need to expand union_types to list
                kwarg_type = T.get_args(kwarg_type)
        else:
            # 3. Do nothing, cannot assign None, as None is a valid type
            continue
        mapping[key] = kwarg_type
    return mapping


def transform_module_inputs(in_module, **kwargs):
    """
    Transform the inputs to all functions in a module.

    Parameters
    ----------
    in_module : Module
        Module containing funcitons which are to be wrapped with transform_function_inputs
    kwarg_types : Dict[str: type]
        Mapping of accepted object types for each arg/kwarg
    convert_types : Tuple[type]
        List of data-types to try to convert, this can be useful when the function is versitile and can
        accept a large number of data-types, hence only a small number of types should be converted.
    kwargs: Any
        Any other kwargs accepted by transform_function_inputs

    Returns
    -------
    Module
        Version of in_module where all functions are wrapped with transform_modul_inputs
    """
    # wrapped_module must be different to original to prevent overriding cached module
    wrapped_module = types.ModuleType(__name__)
    for name in dir(in_module):
        func = getattr(in_module, name)
        # Wrap any functions that are not hidden
        if not name.startswith("_") and isinstance(func, types.FunctionType):
            setattr(
                wrapped_module,
                name,
                transform_function_inputs(func, **kwargs),
            )
        else:
            # If not a func, we just copy
            setattr(wrapped_module, name, func)

    return wrapped_module



import logging
import os

import earthkit.data as data
from earthkit.data.core import Base
from earthkit.data.decorators import locked
from earthkit.data.wrappers import _helpers

LOG = logging.getLogger(__name__)


class Translator(Base):
    def __call__(self):
        """
        Return unmodified data.
        """
        return self.data


_TRANSLATORS = {}


@locked
def _translators():
    return _helpers(
        "translator", _TRANSLATORS, here=os.path.dirname(__file__), package=__name__
    )


def get_translator(source, cls, *args, **kwargs):
    """
    Get the appropriate translator for the source based on the target cls.
    """
    if not isinstance(source, Base):
        source = data.from_object(source)

    for name, h in _translators().items():
        translator = h(source, cls, *args, **kwargs)
        if translator is not None:
            return translator

    raise ValueError(f"Cannot find a translator for class {cls.__name__}")


def transform(*args, **kwargs):
    """
    Executing wrapper for the get_translator class method
    """
    return get_translator(*args, **kwargs)()
